@@ -1,6 +1,6 @@
 /*
  * C++ BMP library.
- * Copyright (C) 2006 - 2011 René Rebe, ExactCODE GmbH Germany
+ * Copyright (C) 2006 - 2016 René Rebe, ExactCODE GmbH Germany
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -17,11 +17,9 @@
  *
  * Lossy based on (in the past more so, but more and more parts got
  * rewritten):
- *
  * Project:  libtiff tools
  * Purpose:  Convert Windows BMP files in TIFF.
  * Author:   Andrey Kiselev, dron@remotesensing.org
- *
  */
 
 #include <iostream>
@@ -32,34 +30,15 @@
 
 #include <string.h>
 #include <ctype.h>
+#include <vector>
 
 #include <Endianess.hh>
-#include <inttypes.h>
-
-static int last_bit_set (int v) {
-  unsigned int i;
-  for (i = sizeof (int) * 8 - 1; i > 0; --i) {
-    if (v & (1L << i))
-      return i;
-  }
-  return 0;
-}
-
-#ifndef O_BINARY
-# define O_BINARY 0
-#endif
+#include <stdint.h>
+#include <Bits.hh>
 
 using Exact::EndianessConverter;
 using Exact::LittleEndianTraits;
 
-enum BMPType
-  {
-    BMPT_WIN4,      /* BMP used in Windows 3.0/NT 3.51/95 */
-    BMPT_WIN5,      /* BMP used in Windows NT 4.0/98/Me/2000/XP */
-    BMPT_OS21,      /* BMP used in OS/2 PM 1.x */
-    BMPT_OS22       /* BMP used in OS/2 PM 2.x */
-  };
-
 /*
  * Bitmap file consists of a BMPFileHeader structure followed by a
  * BMPInfoHeader structure. An array of BMPColorEntry structures (also called
@@ -83,16 +62,21 @@ enum BMPType
 
 enum BMPComprMethod
   {
-    BMPC_RGB = 0L,          /* Uncompressed */
-    BMPC_RLE8 = 1L,         /* RLE for 8 bpp images */
-    BMPC_RLE4 = 2L,         /* RLE for 4 bpp images */
-    BMPC_BITFIELDS = 3L,    /* Bitmap is not compressed and the colour table
+    BMPC_RGB = 0,          /* Uncompressed */
+    BMPC_RLE8 = 1,         /* RLE for 8 bpp images */
+    BMPC_RLE4 = 2,         /* RLE for 4 bpp images */
+    BMPC_BITFIELDS = 3,    /* Bitmap is not compressed and the colour table
 			     * consists of three DWORD color masks that specify
 			     * the red, green, and blue components of each
 			     * pixel. This is valid when used with
 			     * 16- and 32-bpp bitmaps. */
-    BMPC_JPEG = 4L,         /* Indicates that the image is a JPEG image. */
-    BMPC_PNG = 5L           /* Indicates that the image is a PNG image. */
+    BMPC_JPEG = 4,         /* Indicates that the image is a JPEG image. */
+    BMPC_PNG = 5,           /* Indicates that the image is a PNG image. */
+    BMPC_ALPHABITFIELDS = 6,
+    BMPC_CMYK = 11,
+    BMPC_CMYKRLE = 12,
+    BMPC_CMYRTLE = 13,
+    
   };
 
 enum BMPLCSType                 /* Type of logical color space. */
@@ -101,7 +85,7 @@ enum BMPLCSType                 /* Type of logical color space. */
 				 * gamma values are given in the appropriate
 				 * fields. */
     BMPLT_DEVICE_RGB = 1,
-    BMPLT_DEVICE_CMYK = 2
+    BMPLT_DEVICE_CMYK = 2,
   };
 
 #ifdef _MSC_VER
@@ -143,6 +127,7 @@ __attribute__((packed))
 
 struct BMPInfoHeader
 {
+  // Fields used for bitmaps, compatible with Windows NT 3.51 and earlier.
   EndianessConverter<uint32_t,LittleEndianTraits> iSize; /* Size of BMPInfoHeader structure in bytes.
 				 * Should be used to determine start of the
 				 * colour table */
@@ -167,19 +152,18 @@ struct BMPInfoHeader
 				 * unsigned for proper shifting. */
   EndianessConverter<int32_t,LittleEndianTraits> iClrImportant;  /* Number of important colours. If 0, all
 			 * colours are required */
-
-  /*
-   * Fields above should be used for bitmaps, compatible with Windows NT 3.51
-   * and earlier. Windows 98/Me, Windows 2000/XP introduces additional fields:
-   */
-
-  EndianessConverter<int32_t,LittleEndianTraits> iRedMask;       /* Colour mask that specifies the red component
+  // 40 bytes
+  
+  // Windows 98/Me, Windows 2000/XP introduces additional fields:
+  EndianessConverter<uint32_t,LittleEndianTraits> iRedMask;       /* Colour mask that specifies the red component
 			 * of each pixel, valid only if iCompression
 			 * is set to BI_BITFIELDS. */
-  EndianessConverter<int32_t,LittleEndianTraits> iGreenMask;     /* The same for green component */
-  EndianessConverter<int32_t,LittleEndianTraits> iBlueMask;      /* The same for blue component */
-  EndianessConverter<int32_t,LittleEndianTraits> iAlphaMask;     /* Colour mask that specifies the alpha
+  EndianessConverter<uint32_t,LittleEndianTraits> iGreenMask;     /* The same for green component */
+  EndianessConverter<uint32_t,LittleEndianTraits> iBlueMask;      /* The same for blue component */
+  EndianessConverter<uint32_t,LittleEndianTraits> iAlphaMask;     /* Colour mask that specifies the alpha
 			 * component of each pixel. */
+  // 56 bytes
+  
   EndianessConverter<uint32_t,LittleEndianTraits> iCSType;        /* Colour space of the DIB. */
   BMPCIEXYZTriple sEndpoints; /* This member is ignored unless the iCSType
 			       * member specifies BMPLT_CALIBRATED_RGB. */
@@ -190,70 +174,64 @@ struct BMPInfoHeader
 			 * in 16^16 format. */
   EndianessConverter<int32_t,LittleEndianTraits> iGammaGreen;    /* Toned response curve for green. */
   EndianessConverter<int32_t,LittleEndianTraits> iGammaBlue;     /* Toned response curve for blue. */
+  // 108 bytes
 }
 #ifdef __GNUC__
 __attribute__((packed))
 #endif
 ;
 
-/*
- * Info header size in bytes:
- */
-static const unsigned int BIH_WIN4SIZE = 40; /* for BMPT_WIN4 */
-static const unsigned int BIH_WIN5SIZE = 57; /* for BMPT_WIN5 */
-static const unsigned int BIH_OS21SIZE = 12; /* for BMPT_OS21 */
-static const unsigned int BIH_OS22SIZE = 64; /* for BMPT_OS22 */
+// Info header size in bytes:
+static const unsigned BIH_OS21SIZE = 12; // OS/2, Windows 2
+static const unsigned BIH_OS22SIZE = 64;
+static const unsigned BIH_V1SIZE = 40; // Win NT, 3.1
+static const unsigned BIH_V2SIZE = 52;
+static const unsigned BIH_V3SIZE = 56;
+static const unsigned BIH_V4SIZE = 108; // Win NT 4.0, 95
+static const unsigned BIH_V5SIZE = 124; // Win NT 5.0, 98
 
-/*
- * We will use plain byte array instead of this structure, but declaration
- * provided for reference
- */
+// We will use plain byte array instead of this structure, for reference:
 struct BMPColorEntry
 {
   char       bBlue;
   char       bGreen;
   char       bRed;
-  char       bReserved;      /* Must be 0 */
+  char       bReserved; // Must be 0
 };
 
 #ifdef _MSC_VER
 #pragma pack(pop)
 #endif
 
-/*
- * Image data in BMP file stored in BGR (or ABGR) format. We rearrange
- * pixels to RGB (RGBA) format.
- */
+// Image data in BMP file stored in BGR (or ABGR) format:
 static void rearrangePixels(uint8_t* buf, uint32_t width, uint32_t bit_count)
 {
-  char tmp;
-  
   switch (bit_count) {
-    
-  case 16:    /* FIXME: need a sample file */
+  case 16:    // FIXME: need a sample file
     break;
     
   case 24:
-    for (uint32_t i = 0; i < width; i++, buf += 3) {
-      tmp = *buf;
-      *buf = *(buf + 2);
-      *(buf + 2) = tmp;
+    for (uint32_t i = 0; i < width; i++, buf += 3)
+      std::swap(buf[0], buf[2]);
+    break;
+
+  case 48:
+    {
+      uint16_t* buf16 = (uint16_t*)buf;
+      for (uint32_t i = 0; i < width; i++, buf16 += 3)
+	std::swap(buf16[0], buf16[2]);
     }
     break;
-  
+
   case 32:
     {
-      uint8_t* buf1 = buf;
-      for (uint32_t i = 0; i < width; i++, buf += 4) {
-	tmp = *buf;
-	*buf1++ = *(buf + 2);
-	*buf1++ = *(buf + 1);
-	*buf1++ = tmp;
-      }
+      for (uint32_t i = 0; i < width; i++, buf += 4)
+	std::swap(buf[0], buf[2]);
     }
     break;
     
   default:
+    // sub-byte color table rows remain unchanged
     break;
   }
 }
@@ -268,90 +246,61 @@ int BMPCodec::readImage (std::istream* stream, Image& image, const std::string&
     return false;
   }
   
-  /* -------------------------------------------------------------------- */
-  /*      Read the BMPFileHeader. We need iOffBits value only             */
-  /* -------------------------------------------------------------------- */
+  // Read the BMPFileHeader. We need iOffBits value only.
   stream->seekg (10);
   stream->read ((char*)&file_hdr.iOffBits, 4);
   
   // fix the iSize, in early BMP file this is pure garbage
   stream->seekg (0, std::ios::end);
-  file_hdr.iSize = stream->tellg (); // TODO: minus the header?
+  file_hdr.iSize = stream->tellg(); // TODO: minus the header?
 
   int i = readImageWithoutFileHeader(stream, image, decompres, &file_hdr);
   return i;
 }
 
-int BMPCodec::readImageWithoutFileHeader (std::istream* stream, Image& image, const std::string& decompres, BMPFileHeader* file_hdr)
+int BMPCodec::readImageWithoutFileHeader (std::istream* stream, Image& image, const std::string& decompres, BMPFileHeader* _file_hdr)
 {
-  BMPFileHeader file_header; // only used if no file_hdr is supplied
-  BMPInfoHeader info_hdr;
-  enum BMPType bmp_type;
+  BMPFileHeader* file_hdr = _file_hdr;
+  BMPFileHeader file_header = {}; // only used if no file_hdr is supplied
+  BMPInfoHeader info_hdr = {};
   int offset = file_hdr ? sizeof(*file_hdr) : 0;
-
+  
   uint32_t clr_tbl_size = 0, n_clr_elems = 3;
   uint8_t* clr_tbl = 0;
-  uint8_t* data = 0;
-
-  /* -------------------------------------------------------------------- */
-  /*      Read the BMPInfoHeader.                                         */
-  /* -------------------------------------------------------------------- */
   
+  // Read the BMPInfoHeader
   stream->seekg (offset);
   stream->read ((char*)&info_hdr.iSize, 4);
   
-   if (!file_hdr) {
+  if (!_file_hdr) {
     offset = 0;
     file_hdr = &file_header;
     stream->seekg (0, std::ios::end);
-    file_hdr->iSize = stream->tellg ();
+    file_hdr->iSize = stream->tellg();
     file_hdr->iOffBits = info_hdr.iSize; // assumed to follow after info header
-
+    
     stream->seekg (offset + 4);
   }
   
-  if (info_hdr.iSize == BIH_WIN4SIZE)
-    bmp_type = BMPT_WIN4;
-  else if (info_hdr.iSize == BIH_OS21SIZE)
-    bmp_type = BMPT_OS21;
-  else if (info_hdr.iSize == BIH_OS22SIZE || info_hdr.iSize == 16)
-    bmp_type = BMPT_OS22;
-  else
-    bmp_type = BMPT_WIN5;
-  
-  if (bmp_type == BMPT_WIN4 || bmp_type == BMPT_WIN5 ||
-      bmp_type == BMPT_OS22) {
-    stream->read((char*)&info_hdr.iWidth, 4);
-    stream->read((char*)&info_hdr.iHeight, 4);
-    stream->read((char*)&info_hdr.iPlanes, 2);
-    stream->read((char*)&info_hdr.iBitCount, 2);
-    stream->read((char*)&info_hdr.iCompression, 4);
-    stream->read((char*)&info_hdr.iSizeImage, 4);
-    stream->read((char*)&info_hdr.iXPelsPerMeter, 4);
-    stream->read((char*)&info_hdr.iYPelsPerMeter, 4);
-    stream->read((char*)&info_hdr.iClrUsed, 4);
-    stream->read((char*)&info_hdr.iClrImportant, 4);
-    stream->read((char*)&info_hdr.iRedMask, 4);
-    stream->read((char*)&info_hdr.iGreenMask, 4);
-    stream->read((char*)&info_hdr.iBlueMask, 4);
-    stream->read((char*)&info_hdr.iAlphaMask, 4);
-    
-    n_clr_elems = 4;
-    image.setResolution((2.54 * info_hdr.iXPelsPerMeter + .05) / 100,
-                        (2.54 * info_hdr.iYPelsPerMeter + .05) / 100);
+  unsigned iSize = info_hdr.iSize;
+  switch (iSize) {
+  case 16:
+    iSize = BIH_OS22SIZE; break;
+  case BIH_OS21SIZE:
+  case BIH_OS22SIZE:
+  case BIH_V1SIZE:
+  case BIH_V2SIZE:
+  case BIH_V3SIZE:
+  case BIH_V4SIZE:
+  case BIH_V5SIZE:
+    break;
+  default:
+    std::cerr << "Unknown header size: " << info_hdr.iSize << std::endl;
+    return false;
   }
   
-  if (bmp_type == BMPT_OS22) {
-    /* 
-     * FIXME: different info in different documents
-     * regarding this!
-     */
-    n_clr_elems = 3;
-  }
-
-  if (bmp_type == BMPT_OS21) {
+  if (iSize == BIH_OS21SIZE) {
     int16_t  iShort;
-    
     stream->read ((char*)&iShort, 2);
     info_hdr.iWidth = iShort;
     stream->read ((char*)&iShort, 2);
@@ -363,44 +312,73 @@ int BMPCodec::readImageWithoutFileHeader (std::istream* stream, Image& image, co
     info_hdr.iCompression = BMPC_RGB;
     n_clr_elems = 3;
   }
+  else{
+    stream->read((char*)&info_hdr.iWidth, 4);
+    stream->read((char*)&info_hdr.iHeight, 4);
+    stream->read((char*)&info_hdr.iPlanes, 2);
+    stream->read((char*)&info_hdr.iBitCount, 2);
+    stream->read((char*)&info_hdr.iCompression, 4);
+    stream->read((char*)&info_hdr.iSizeImage, 4);
+    stream->read((char*)&info_hdr.iXPelsPerMeter, 4);
+    stream->read((char*)&info_hdr.iYPelsPerMeter, 4);
+    stream->read((char*)&info_hdr.iClrUsed, 4);
+    stream->read((char*)&info_hdr.iClrImportant, 4);
+    n_clr_elems = 4;
+    
+    if (iSize >= BIH_V2SIZE ||
+	info_hdr.iCompression == BMPC_BITFIELDS ||
+	info_hdr.iCompression == BMPC_ALPHABITFIELDS) {
+      stream->read((char*)&info_hdr.iRedMask, 4);
+      stream->read((char*)&info_hdr.iGreenMask, 4);
+      stream->read((char*)&info_hdr.iBlueMask, 4);
+      if (iSize >= BIH_V3SIZE || info_hdr.iCompression == BMPC_ALPHABITFIELDS)
+	stream->read((char*)&info_hdr.iAlphaMask, 4);
+      
+      /*std::cerr << std::hex << "red mask: " << info_hdr.iRedMask
+		  << ", green mask: " << info_hdr.iGreenMask
+		  << ", blue mask: " << info_hdr.iBlueMask
+		  << ", alpha mask: " << info_hdr.iAlphaMask << std::dec << std::endl;
+      */
+    }
+  }
   
-  switch (info_hdr.iBitCount) {
-  case 1:
-    //  case 2:
-  case 4:
-  case 8:
-  case 16:
-  case 24:
-  case 32:
-    break;
-  default:
-    std::cerr << "BMPCodec:: Cannot read " << info_hdr.iBitCount
-	      << " bit files." << std::endl;
-    return false;
+  //std::cerr << "size: " << iSize << ", bits: " << info_hdr.iBitCount << ", compr: " << info_hdr.iCompression << std::endl;
+  
+  if (iSize == BIH_OS22SIZE) {
+    //FIXME: different info in different documents regarding this!
+    n_clr_elems = 3;
   }
 
+  image.setResolution((2.54 * info_hdr.iXPelsPerMeter) / 100 + .5,
+		      (2.54 * info_hdr.iYPelsPerMeter) / 100 + .5);
+  
   image.w = info_hdr.iWidth;
   image.h = std::abs(info_hdr.iHeight); // negative when upside-down
   
   switch (info_hdr.iBitCount)
     {
     case 1:
-    case 2:
+    case 2: // valid in Win CE?
     case 4:
     case 8:
       image.spp = 1;
       image.bps = info_hdr.iBitCount;
 
-      /* Allocate memory for colour table and read it. */
+      // Allocate memory for colour table and read it.
       if (info_hdr.iClrUsed)
-	clr_tbl_size = ((uint32_t)(1 << image.bps) < info_hdr.iClrUsed) ?
-	  1 << image.bps : info_hdr.iClrUsed;
+	clr_tbl_size =
+	  ((uint32_t)(1 << image.bps) < info_hdr.iClrUsed) ? 1 << image.bps : info_hdr.iClrUsed;
       else
 	clr_tbl_size = 1 << image.bps;
 
       //std::cerr << "n_clr_elems: " << n_clr_elems << ", clr_tbl_size: " << clr_tbl_size << std::endl;
 
-      clr_tbl = (uint8_t *) malloc (n_clr_elems * clr_tbl_size);
+      // if we had no file_hdr, update offset to compensate for color table
+      if (!_file_hdr) {
+        file_hdr->iOffBits = file_hdr->iOffBits + n_clr_elems * clr_tbl_size;
+      }
+
+      clr_tbl = (uint8_t*)malloc(n_clr_elems * clr_tbl_size);
       if (!clr_tbl) {
 	std::cerr << "BMPCodec:: Can't allocate space for color table" << std::endl;
 	return false;
@@ -410,33 +388,37 @@ int BMPCodec::readImageWithoutFileHeader (std::istream* stream, Image& image, co
       stream->read ((char*)clr_tbl, n_clr_elems * clr_tbl_size);
       
       /*for(clr = 0; clr < clr_tbl_size; ++clr) {
-	printf ("%d: r: %d g: %d b: %d\n",
-		clr,
-		clr_tbl[clr*n_clr_elems+2],
-		clr_tbl[clr*n_clr_elems+1],
-		clr_tbl[clr*n_clr_elems]);
+	printf ("%d: r: %d g: %d b: %d\n", clr,
+		clr_tbl[clr*n_clr_elems+2], clr_tbl[clr*n_clr_elems+1], clr_tbl[clr*n_clr_elems]);
       }*/
       break;
     
     case 16:
     case 24:
-      image.spp = 3;
-      image.bps = info_hdr.iBitCount / image.spp;
-      break;
-    
     case 32:
       image.spp = 3;
       image.bps = 8;
+      
+      if (info_hdr.iCompression == BMPC_ALPHABITFIELDS)
+	image.spp = 4;
+      else if (iSize >= BIH_V2SIZE && info_hdr.iCompression == BMPC_BITFIELDS && info_hdr.iAlphaMask != 0)
+	image.spp = 4; // TODO: does gray + alpha exist?
+      else if (info_hdr.iCompression == BMPC_RGB && info_hdr.iBitCount == 32)
+	image.spp = 4;
+      
       break;
     
-    default:
+    case 48:
+      image.spp = 3;
+      image.bps = 16;
       break;
+
+    default:
+      std::cerr << "BMPCodec:: Cannot read " << info_hdr.iBitCount
+		<< " bit files." << std::endl;
+      return false;
     }
   
-  uint32_t stride = image.stride ();
-  /*printf ("w: %d, h: %d, spp: %d, bps: %d, colorspace: %d\n",
-   *w, *h, *spp, *bps, info_hdr.iCompression); */
-  
   // detect old style bitmask images
   if (info_hdr.iCompression == BMPC_RGB && info_hdr.iBitCount == 16)
     {
@@ -445,17 +427,17 @@ int BMPCodec::readImageWithoutFileHeader (std::istream* stream, Image& image, co
       info_hdr.iBlueMask = 0x1f;
       info_hdr.iGreenMask = 0x1f << 5;
       info_hdr.iRedMask = 0x1f << 10;
+      info_hdr.iAlphaMask = 0;
     }
   
-  /* -------------------------------------------------------------------- */
-  /*  Read uncompressed image data.                                       */
-  /* -------------------------------------------------------------------- */
-
+  unsigned stride = image.stride();
+  
+  // Read uncompressed image data.
   switch (info_hdr.iCompression) {
   case BMPC_BITFIELDS:
-    // we unpack bitfields to plain RGB
-    image.bps = 8;
-    stride = image.stride ();
+  case BMPC_ALPHABITFIELDS:
+    image.bps = 8; // we unpack bitfields to plain RGB
+    stride = image.stride();
     
   case BMPC_RGB:
     {
@@ -463,180 +445,173 @@ int BMPCodec::readImageWithoutFileHeader (std::istream* stream, Image& image, co
       
       /*std::cerr << "bitcount: " << info_hdr.iBitCount << ", stride: " << stride
                 << ", file stride: " << file_stride << std::endl;
-     
-      std::cerr << std::hex << "red mask: " << info_hdr.iRedMask
-                << ", green mask: " << info_hdr.iGreenMask
-                << ", blue mask: " << info_hdr.iBlueMask << std::dec << std::endl; */
+      if (info_hdr.iCompression == BMPC_BITFIELDS || info_hdr.iCompression == BMPC_ALPHABITFIELDS)
+	std::cerr << std::hex << "red mask: " << info_hdr.iRedMask
+		  << ", green mask: " << info_hdr.iGreenMask
+		  << ", blue mask: " << info_hdr.iBlueMask
+		  << ", alpha mask: " << info_hdr.iAlphaMask << std::dec << std::endl;
+      */
+      if (file_stride > stride)
+	stride = file_stride; // use the BMP's native stride
+      image.resize(image.w, image.h, stride);
+      
+      const int r_bits = Exact::popcountf(info_hdr.iRedMask);
+      const int r_shift = Exact::ms_bit_set(info_hdr.iRedMask) + 1 - r_bits;
+      const int g_bits = Exact::popcountf(info_hdr.iGreenMask);
+      const int g_shift = Exact::ms_bit_set(info_hdr.iGreenMask) + 1 - g_bits;
+      const int b_bits = Exact::popcountf(info_hdr.iBlueMask);
+      const int b_shift = Exact::ms_bit_set(info_hdr.iBlueMask) + 1 - b_bits;
+      const int a_bits = Exact::popcountf(info_hdr.iAlphaMask);
+      const int a_shift = Exact::ms_bit_set(info_hdr.iAlphaMask) + 1 - a_bits;
 
-      image.resize(image.w, image.h);
-      data = image.getRawData();
-      uint8_t* row_data = (uint8_t*) malloc (file_stride);
-      if (!row_data) {
-	std::cerr << "Can't allocate space for image buffer\n";
-	goto bad1;
-      }
-      
-      const int r_shift = last_bit_set (info_hdr.iRedMask) - 7;
-      const int g_shift = last_bit_set (info_hdr.iGreenMask) - 7;
-      const int b_shift = last_bit_set (info_hdr.iBlueMask) - 7;
-      
+      /*std::cerr << r_bits << " " << r_shift << " " << g_bits << " " << g_shift << " "
+		<< b_bits << " " << b_shift << " " << a_bits << " " << a_shift << std::endl;
+      */
+      uint8_t* data = image.getRawData();
       for (uint32_t row = 0; row < (uint32_t)image.h; ++row)
       {
 	std::istream::pos_type offset = file_hdr->iOffBits + row * file_stride;
-	stream->seekg (offset);
+	stream->seekg(offset);
 	
-	if (stream->tellg () != offset) {
+	if (stream->tellg() != offset) {
 	  std::cerr << "scanline " << row << " Seek error: " << stream->tellg() << " vs " << offset << std::endl;
 	}
 	
-	if (stream->read ((char*)row_data, file_stride) < 0) {
-	  std::cerr << "scanline " << row << ": Read error\n";
-	}
+	uint8_t* row_ptr = data +
+	  stride * (info_hdr.iHeight < 0 ? row : image.h - row - 1);
 	
-	// convert to RGB
-	uint8_t* rgb_ptr = data + stride * (info_hdr.iHeight < 0 ? row : image.h - row - 1);
-
-	if (info_hdr.iCompression == BMPC_BITFIELDS)
+	stream->read((char*)row_ptr, file_stride);
+	if (!stream->good()) {
+	  std::cerr << "bmp read error: scanline " << row << "\n";
+	} else {
+	  // convert to RGB
+	  if (info_hdr.iCompression == BMPC_BITFIELDS || info_hdr.iCompression == BMPC_ALPHABITFIELDS)
 	  {
-	    uint8_t* bf_ptr = row_data;
+	    const int spp = image.spp;
+	    int beg = 0, end = image.w; int8_t inc = 1;
+	    if (image.spp * image.bps > info_hdr.iBitCount) {
+	      // reverse, not to clobber in-line data
+	      beg = image.w - 1;
+	      end = -1;
+	      inc = -1;
+	    }
 	    
-	    for (int i = 0; i < image.w; ++i, rgb_ptr += 3)
+	    //if (row == 0) std::cerr << beg << " " << end << " " << (int)inc << std::endl;
+	    for (int i = beg; i != end; i += inc)
 	      {
-		int val = 0;
-		for (int bits = 0; bits < info_hdr.iBitCount; bits += 8)
+		uint8_t* bf_ptr = row_ptr + i * info_hdr.iBitCount / 8;
+		int32_t val = *bf_ptr++; // 1st 8 bits
+		for (int bits = 8; bits < info_hdr.iBitCount; bits += 8)
 		  val |= (*bf_ptr++) << bits;
 		    
-		if (r_shift > 0)
-		  rgb_ptr[0] = (val & info_hdr.iRedMask) >> r_shift;
-		else
-		  rgb_ptr[0] = (val & info_hdr.iRedMask) << -r_shift;
-		if (g_shift > 0)
-		  rgb_ptr[1] = (val & info_hdr.iGreenMask) >> g_shift;
-		else
-		  rgb_ptr[1] = (val & info_hdr.iGreenMask) << -g_shift;
-		if (b_shift > 0)
-		  rgb_ptr[2] = (val & info_hdr.iBlueMask) >> b_shift;
-		else
-		  rgb_ptr[2] = (val & info_hdr.iBlueMask) << -b_shift;
+		row_ptr[i*spp + 0] =
+		  ((val & info_hdr.iRedMask) >> r_shift) * 0xff / ((1 << r_bits) - 1);
+		row_ptr[i*spp + 1] =
+		  ((val & info_hdr.iGreenMask) >> g_shift) * 0xff / ((1 << g_bits) - 1);
+		row_ptr[i*spp + 2] =
+		  ((val & info_hdr.iBlueMask) >> b_shift) * 0xff / ((1 << b_bits) - 1);
+		if (spp > 3)
+		  row_ptr[i*spp + 3] = 
+		    ((val & info_hdr.iAlphaMask) >> a_shift) * 0xff / ((1 << a_bits) - 1);
 	      }
+	  } else {
+	    rearrangePixels (row_ptr, image.w, info_hdr.iBitCount);
 	  }
-	else {
-	  rearrangePixels (row_data, image.w, info_hdr.iBitCount);
-	  memcpy (rgb_ptr, row_data, stride);
 	}
       }
-      free(row_data);
     }
     break;
     
-    /* -------------------------------------------------------------------- */
-    /*  Read compressed image data.                                         */
-    /* -------------------------------------------------------------------- */
+    // Read compressed image data
   case BMPC_RLE4:
   case BMPC_RLE8:
     {
-      uint32_t	i, j, k, runlength, x;
-      uint32_t	compr_size, uncompr_size;
-      uint8_t   *comprbuf;
-      uint8_t   *uncomprbuf;
+      const unsigned compr_size = file_hdr->iSize - file_hdr->iOffBits;
+      const unsigned uncompr_size = image.w * image.h;
       
-      //std::cerr << "RLE" << (info_hdr.iCompression == BMPC_RLE4 ? "4" : "8")
-      //	<< " compressed\n";
+      image.bps = 8;
+      image.resize(image.w, image.h);
+      uint8_t* uncomprbuf = image.getRawData();
+      memset(uncomprbuf, 0, uncompr_size); // for skipped pixels, ...
       
-      compr_size = file_hdr->iSize - file_hdr->iOffBits;
-      uncompr_size = image.w * image.h;
+#ifdef _MSC_VER
+      std::vector<uint8_t> comprbuf(compr_size);
+#else
+      uint8_t comprbuf[compr_size];
+#endif
+      stream->seekg(*file_hdr->iOffBits);
+      stream->read((char*)&comprbuf[0], compr_size);
       
-      comprbuf = (uint8_t *) malloc( compr_size );
-      if (!comprbuf) {
-	std::cerr << "Can't allocate space for compressed scanline buffer\n";
-	goto bad1;
-      }
-      uncomprbuf = (uint8_t *) malloc( uncompr_size );
-      if (!uncomprbuf) {
-	std::cerr << "Can't allocate space for uncompressed scanline buffer\n";
-	goto bad1;
-      }
-      
-      stream->seekg (*file_hdr->iOffBits);
-      stream->read ((char*)comprbuf, compr_size);
-      i = j = x = 0;
-      
-      while( j < uncompr_size && i < compr_size ) {
-	if ( comprbuf[i] ) {
-	  runlength = comprbuf[i++];
-	  for ( k = 0;
-		runlength > 0 && j < uncompr_size && i < compr_size && x < (uint32_t)image.w;
-		++k, ++x) {
+      int y = image.h - 1;
+      uint8_t* rowptr = uncomprbuf + y * image.w;
+      for (unsigned i = 0, x = 0; y >= 0 && i < compr_size;) {
+	if (comprbuf[i]) {
+	  uint8_t runlength = comprbuf[i++];
+	  for (unsigned k = 0;
+	       runlength > 0 && i < compr_size && x < (uint32_t)image.w;
+	       ++k) {
 	    if (info_hdr.iBitCount == 8)
-	      uncomprbuf[j++] = comprbuf[i];
+	      rowptr[x++] = comprbuf[i];
 	    else {
-	      if ( k & 0x01 )
-		uncomprbuf[j++] = comprbuf[i] & 0x0F;
+	      if (k & 1)
+		rowptr[x++] = comprbuf[i] & 0x0F;
 	      else
-		uncomprbuf[j++] = (comprbuf[i] & 0xF0) >> 4;
+		rowptr[x++] = (comprbuf[i] & 0xF0) >> 4;
 	    }
 	    runlength--;
 	  }
-	  i++;
+	  ++i;
 	} else {
-	  i++;
-	  if ( comprbuf[i] == 0 ) {         /* Next scanline */
-	    i++;
-	    x = 0;;
+	  ++i;
+	  uint8_t v = comprbuf[i];
+	  if (v == 0) { // Next scanline
+	    ++i;
+	    x = 0; --y;
+	    rowptr = uncomprbuf + y * image.w;
 	  }
-	  else if ( comprbuf[i] == 1 )    /* End of image */
+	  else if (v == 1) // End of image
 	    break;
-	  else if ( comprbuf[i] == 2 ) {  /* Move to... */
-	    i++;
-	    if ( i < compr_size - 1 ) {
-	      j += comprbuf[i] + comprbuf[i+1] * image.w;
+	  else if (v == 2) { // Move to...
+	    ++i;
+	    if (i < compr_size - 1) {
+	      x += comprbuf[i];
+	      y -= comprbuf[i+1];
+	      rowptr = uncomprbuf + y * image.w;
 	      i += 2;
 	    }
 	    else
 	      break;
-	  } else {                         /* Absolute mode */
-	    runlength = comprbuf[i++];
-	    for ( k = 0; k < runlength && j < uncompr_size && i < compr_size; k++, x++)
+	  } else { // uncompressed mode
+	    uint8_t runlength = comprbuf[i++];
+	    for (unsigned k = 0; k < runlength && x < image.w && i < compr_size; ++k)
 	      {
 		if (info_hdr.iBitCount == 8)
-		  uncomprbuf[j++] = comprbuf[i++];
+		  rowptr[x++] = comprbuf[i++];
 		else {
-		  if ( k & 0x01 )
-		    uncomprbuf[j++] = comprbuf[i++] & 0x0F;
-		  else
-		    uncomprbuf[j++] = (comprbuf[i] & 0xF0) >> 4;
+		  if (k & 1)
+		    rowptr[x++] = v & 0x0F;
+		  else {
+		    v = comprbuf[i++]; // store for next nibble
+		    rowptr[x++] = (v & 0xF0) >> 4;
+		  }
 		}
 	      }
-	    /* word boundary alignment */
-	    if (info_hdr.iBitCount == 4)
-	      k /= 2;
-	    if ( k & 0x01 )
-	      i++;
+	    
+	    // word boundary alignment
+ 	    if (i & 1)
+	      ++i;
 	  }
 	}
       }
-      
-      free(comprbuf);
-      data = (uint8_t *) malloc( uncompr_size );
-      if (!data) {
-	std::cerr << "Can't allocate space for final uncompressed scanline buffer\n";
-	goto bad1;
-      }
-      
-      // TODO: suboptimal, later improve the above to yield the corrent orientation natively
-      for (uint32_t row = 0; row < (uint32_t)image.h; ++row) {
-	memcpy (data + row * image.w, uncomprbuf + (image.h - 1 - row) * image.w, image.w);
-	rearrangePixels(data + row * image.w, image.w, info_hdr.iBitCount);
-      }
-      
-      free(uncomprbuf);
-      image.bps = 8;
     }
-    image.setRawData(data);
     break;
-  } /* switch */
+    
+  default:
+    std::cerr << "BMPCodec: unsupported compression: " << info_hdr.iCompression << std::endl;
+    return false;
+  }
   
-  // convert to RGB color-space - we do not handle palete images internally
+  // convert to RGB color-space - we do not handle palette images internally
   
   // no color table anyway or RGB* ?
   if (clr_tbl && image.spp < 3)
@@ -661,53 +636,43 @@ int BMPCodec::readImageWithoutFileHeader (std::istream* stream, Image& image, co
       free(clr_tbl);
       clr_tbl = NULL;
     }
+  else if (image.spp == 4 && (iSize < BIH_V2SIZE && info_hdr.iCompression == BMPC_RGB)) {
+    colorspace_rgba8_to_rgb8(image);
+  }
   
   return true;
-  
- bad1:
-  if (clr_tbl)
-    free(clr_tbl);
-  clr_tbl = NULL;
-  
-  return false;
 }
 
 bool BMPCodec::writeImage (std::ostream* stream, Image& image, int quality,
 			   const std::string& compress)
 {
-  if (image.bps > 8 || image.bps == 2 || image.spp > 3) {
-    std::cerr << "BMPCodec: " << image.bps << " bits and "
-	      << image.spp << " samples not supported." << std::endl;
+  const int hdr_size = image.spp == 4 ? BIH_V3SIZE : BIH_V1SIZE;
+  const unsigned stride = image.stride();
+  const int n_clr_elems = 4; // we write "modern" formats, not the vintage OS/2 flavour
+  
+  if (image.bps > 16 || image.spp > 4) {
+    std::cerr << "BMPCodec: " << image.bps << " bits and " << image.spp << " samples not supported." << std::endl;
     return false;
   }
   
-  const int n_clr_elems = 4; // we write "modern" formats, not the first OS2 flavours
-  
-  BMPFileHeader file_hdr;
-  BMPInfoHeader info_hdr;
-  
-  int hdr_size = BIH_WIN4SIZE;
-  int stride = image.stride ();
-  
-  memset (&file_hdr, 0, sizeof (file_hdr));
-  memset (&info_hdr, 0, sizeof (info_hdr));
+  BMPFileHeader file_hdr = {};
+  BMPInfoHeader info_hdr = {};
 
   // BMPFileHeader
-
   file_hdr.bType[0] = 'B';
   file_hdr.bType[1] = 'M';
   
   // BMPInfoHeader
-  
   info_hdr.iSize = hdr_size;
   info_hdr.iWidth = image.w;
   info_hdr.iHeight = image.h;
   info_hdr.iPlanes = 1;
   info_hdr.iBitCount = image.spp * image.bps;
   info_hdr.iCompression = BMPC_RGB;
-  info_hdr.iSizeImage  = image.stride()*image.h; // TODO: compressed size
-  info_hdr.iXPelsPerMeter = (int32_t) (image.resolutionX() * 100 / 2.54);
-  info_hdr.iYPelsPerMeter = (int32_t) (image.resolutionY() * 100 / 2.54);
+  const int file_stride = ((image.w * info_hdr.iBitCount + 7) / 8 + 3) / 4 * 4; // BMP 4byte aligned
+  info_hdr.iSizeImage  = file_stride * image.h; // TODO: compressed size
+  info_hdr.iXPelsPerMeter = (int32_t) (100. * image.resolutionX() / 2.54 + .5);
+  info_hdr.iYPelsPerMeter = (int32_t) (100. * image.resolutionY() / 2.54 + .5);
   info_hdr.iClrUsed = image.spp == 1 ? 1 << image.bps : 0;
   info_hdr.iClrImportant = 0;
   info_hdr.iRedMask = 0;
@@ -715,8 +680,6 @@ bool BMPCodec::writeImage (std::ostream* stream, Image& image, int quality,
   info_hdr.iBlueMask = 0;
   info_hdr.iAlphaMask = 0;
   
-  // BMP image payload needs to be 4 byte aligned :-(
-  int file_stride = ((image.w * info_hdr.iBitCount + 7) / 8 + 3) / 4 * 4;
   
   file_hdr.iOffBits = sizeof(file_hdr) + hdr_size + info_hdr.iClrUsed * n_clr_elems;
   file_hdr.iSize =  file_hdr.iOffBits + file_stride * image.h;
@@ -729,7 +692,7 @@ bool BMPCodec::writeImage (std::ostream* stream, Image& image, int quality,
   if (info_hdr.iClrUsed) {
     int n = info_hdr.iClrUsed;
 #ifdef _MSC_VER
-    std::vector<uint8_t> _clrtbl(n_clr_elems*n);
+    std::vector<uint8_t> _clrtbl(n_clr_elems * n);
     uint8_t* clrtbl = &_clrtbl[0];
 #else
     uint8_t clrtbl [n_clr_elems*n];
@@ -748,17 +711,18 @@ bool BMPCodec::writeImage (std::ostream* stream, Image& image, int quality,
   case BMPC_RGB:
     {
 #ifdef _MSC_VER
-      std::vector<uint8_t> _payload(file_stride);
-      uint8_t* payload = &_payload[0];
+      std::vector<uint8_t> payload(file_stride);
 #else
-      uint8_t payload [file_stride];
+      uint8_t payload[file_stride];
 #endif
-      for (int row = image.h-1; row >=0; --row)
+      for (int i = stride; i < file_stride; ++i)
+	payload[i] = 0; // zero initialize padding
+      for (int row = image.h-1; row >= 0; --row)
 	{
-	  memcpy (payload, image.getRawData() + stride*row, stride);
-	  rearrangePixels (payload, image.w, info_hdr.iBitCount);
+	  memcpy(&payload[0], image.getRawData() + stride * row, stride);
+	  rearrangePixels(&payload[0], image.w, info_hdr.iBitCount);
 	  
-	  if (!stream->write ((char*)payload, file_stride)) {
+	  if (!stream->write((char*)&payload[0], file_stride)) {
 	    std::cerr << "scanline " << row << " write error" << std::endl;
 	    return false;
 	  }
@@ -769,7 +733,7 @@ bool BMPCodec::writeImage (std::ostream* stream, Image& image, int quality,
   default:
     std::cerr << "unsupported compression method writing bmp" << std::endl;
     return false;
-  } /* switch */
+  }
   
   return true;
 }
